@@ -39,17 +39,19 @@ refs will be ignored.
 
 from . import VERSION
 
-from argparse import ArgumentParser
-import atexit
+from argparse import ArgumentParser, Namespace, SUPPRESS
 from collections import OrderedDict
 from configparser import ConfigParser
+import dataclasses
 import fnmatch
 import gi
 import logging
 import os
-import shutil
+from pathlib import Path
+import shlex
 import subprocess
-import tempfile
+from tempfile import TemporaryDirectory
+import yaml
 
 gi.require_version('OSTree', '1.0')
 from gi.repository import GLib, Gio, OSTree  # noqa: E402
@@ -62,6 +64,140 @@ class OTReceiveError(Exception):
     pass
 
 
+class OTReceiveConfigError(OTReceiveError):
+    """Errors from ostree-receive configuration"""
+    pass
+
+
+@dataclasses.dataclass
+class OTReceiveConfig:
+    """OTReceiveRepo configuration
+
+    Configuration can be provided from a file or command line arguments using
+    the load method. Config files are YAML mappings with the option names
+    below using hypens instead of underscores. By default, the paths
+    ~/.config/ostree/ostree-receive.conf and /etc/ostree/ostree-receive.conf
+    are read unless the OSTREE_RECEIVE_CONF environment variable is set. That
+    can be used to point to a file to be read.
+
+    Supported configuration options:
+
+    root: Specify a repo root directory. When None or '', any repo path is
+      allowed and paths are resolved relative to the current working
+      directory. This is typically the user's home directory.
+    gpg_sign: GPG key IDs for signing received commits and repo metadata.
+    gpg_homedir: GnuPG home directory for loading GPG signing keys.
+    update: Update the repo metadata after receiving commits.
+    update_hook: Program to run after new commits have been made. The program
+      will be executed with the environment variable OSTREE_RECEIVE_REPO set
+      to the absolute path of the OSTree repository and the environment
+      variable OSTREE_RECEIVE_REFS set to the set of refs received separated
+      by whitespace.
+    log_level: Set the log level. See the logging module for available levels.
+    force: Force receiving commits even if nothing changed or the remote
+      commits are not newer than the current commits.
+    dry_run: Only show what would be done without making any commits.
+    """
+    root: str = None
+    # It would be nice to make this list[str], but that would break
+    gpg_sign: list = dataclasses.field(default_factory=list)
+    gpg_homedir: str = None
+    update: bool = True
+    update_hook: str = None
+    log_level: str = 'INFO'
+    force: bool = False
+    dry_run: bool = False
+
+    def __post_init__(self):
+        # Validate the instance.
+        for field in dataclasses.fields(self):
+            value = getattr(self, field.name)
+
+            # Validate the type. None is allowed if the default is None.
+            if value is None and field.default is None:
+                continue
+            elif not isinstance(value, field.type):
+                inst_type = type(value)
+                raise OTReceiveConfigError(
+                    f'{field.name} must be an instance of '
+                    f'{field.type}, but found {inst_type}'
+                )
+
+    @classmethod
+    def load(cls, paths=None, args=None):
+        """Create instance from config files and arguments
+
+        If paths is None, default_paths() will be used.
+        """
+        conf = {}
+        fields = {field.name for field in dataclasses.fields(cls)}
+        if paths is None:
+            paths = cls.default_paths()
+
+        # Load config file options
+        for p in paths:
+            try:
+                path = Path(p).expanduser().resolve()
+            except TypeError as err:
+                raise OTReceiveConfigError(err) from None
+            if not path.exists():
+                logger.debug('Skipping missing config file %s', path)
+                continue
+
+            logger.debug('Loading config file %s', path)
+            with path.open() as f:
+                data = yaml.safe_load(f)
+            if data is None:
+                logger.debug('Ignoring empty config file %s', path)
+                continue
+            elif not isinstance(data, dict):
+                raise OTReceiveConfigError(
+                    f'Config file {path} is not a YAML mapping'
+                )
+
+            for option, value in data.items():
+                if option not in fields:
+                    logger.warning(
+                        'Unrecognized option %s in config file %s',
+                        option, path
+                    )
+                    continue
+                logger.debug(
+                    'Setting option %s to %s from %s', option, value, path
+                )
+                conf[option] = value
+
+        # Load argument options
+        if args is not None:
+            if not isinstance(args, Namespace):
+                raise OTReceiveConfigError(
+                    'args is not an argparse.Namespace instance'
+                )
+
+            logger.debug('Loading arguments %s', args)
+            for arg, value in vars(args).items():
+                if arg not in fields:
+                    logger.debug('Ignoring argument %s', arg)
+                    continue
+                logger.debug('Setting option %s to %s from args', arg, value)
+                conf[arg] = value
+
+        return cls(**conf)
+
+    @staticmethod
+    def default_paths():
+        """Return list of default configuration files"""
+        env_config = os.getenv('OSTREE_RECEIVE_CONF')
+        if env_config:
+            return [Path(env_config)]
+
+        config_home = Path(os.getenv('XDG_CONFIG_HOME', '~/.config'))
+        return [
+            Path('/etc/ostree/ostree-receive.conf'),
+            config_home / 'ostree/ostree-receive.conf',
+        ]
+
+
 class OTReceiveRepo(OSTree.Repo):
     # The fake remote name
     REMOTE_NAME = '_receive'
@@ -73,17 +209,40 @@ class OTReceiveRepo(OSTree.Repo):
         OSTree.REPO_METADATA_REF,
     )
 
-    def __init__(self, path, url):
-        self.path = path
+    def __init__(self, path, url, config=None):
+        self.path = Path(path)
         self.url = url
         self.remotes_dir = None
-        self._exit_func = atexit.register(self.cleanup)
+
+        if config:
+            if not isinstance(config, OTReceiveConfig):
+                raise OTReceiveError(
+                    'config is not an OTReceiveConfig instance'
+                )
+            self.config = config
+        else:
+            self.config = OTReceiveConfig()
+
+        if self.config.root:
+            repo_root = Path(self.config.root).resolve()
+            if not self.path.is_absolute():
+                # Join the relative path to the root.
+                self.path = repo_root.joinpath(self.path)
+
+            # Make sure the path is below the root.
+            self.path = self.path.resolve()
+            try:
+                self.path.relative_to(repo_root)
+            except ValueError:
+                raise OTReceiveError(f'repo {path} not found') from None
+
+        logger.debug('Using repo path %s', self.path)
 
         # Create a temporary remote config file. Just an empty URL is
         # needed and the rest of the parameters will be supplied in the
         # pull options.
-        self.remotes_dir = tempfile.mkdtemp(prefix='ostree-receive-')
-        remote_config_path = os.path.join(self.remotes_dir,
+        self.remotes_dir = TemporaryDirectory(prefix='ostree-receive-')
+        remote_config_path = os.path.join(self.remotes_dir.name,
                                           f'{self.REMOTE_NAME}.conf')
         remote_config = ConfigParser()
         remote_section = f'remote "{self.REMOTE_NAME}"'
@@ -94,14 +253,12 @@ class OTReceiveRepo(OSTree.Repo):
         with open(remote_config_path, 'w') as f:
             remote_config.write(f, space_around_delimiters=False)
 
-        repo_file = Gio.File.new_for_path(path)
-        super().__init__(path=repo_file, remotes_config_dir=self.remotes_dir)
+        repo_file = Gio.File.new_for_path(os.fspath(self.path))
+        super().__init__(path=repo_file,
+                         remotes_config_dir=self.remotes_dir.name)
         self.open()
 
     def __enter__(self):
-        # Unregister the exit function since the cleanup will be done
-        # when exiting the context.
-        atexit.unregister(self._exit_func)
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
@@ -110,10 +267,11 @@ class OTReceiveRepo(OSTree.Repo):
     def cleanup(self):
         """Cleanup instance temporary directory
 
-        This will be called automatically when the program or context exits.
+        This will be called automatically when the instance is deleted
+        or the context exits.
         """
         if self.remotes_dir:
-            shutil.rmtree(self.remotes_dir)
+            self.remotes_dir.cleanup()
             self.remotes_dir = None
 
     def _get_commit_timestamp(self, rev):
@@ -143,7 +301,7 @@ class OTReceiveRepo(OSTree.Repo):
         finally:
             progress.finish()
 
-    def copy_commit(self, rev, ref, force=False):
+    def copy_commit(self, rev, ref):
         _, src_variant, src_state = self.load_commit(rev)
         if src_state != OSTree.RepoCommitState.NORMAL:
             raise OTReceiveError(f'Cannot copy irregular commit {rev}')
@@ -209,6 +367,12 @@ class OTReceiveRepo(OSTree.Repo):
                                                          commit_root,
                                                          commit_time)
 
+        for key in self.config.gpg_sign:
+            logger.debug('Signing commit %s with key %s',
+                         commit_checksum, key)
+            self.sign_commit(commit_checksum, key,
+                             self.config.gpg_homedir)
+
         # Update the ref
         self.transaction_set_refspec(ref, commit_checksum)
 
@@ -233,14 +397,47 @@ class OTReceiveRepo(OSTree.Repo):
         return any(filter(self._is_flatpak_ref, refs))
 
     def update_repo_metadata(self):
+        sign_opts = []
+        if self.config.gpg_sign:
+            sign_opts += [f'--gpg-sign={key}' for key in self.config.gpg_sign]
+            if self.config.gpg_homedir:
+                sign_opts.append(f'--gpg-homedir={self.config.gpg_homedir}')
         if self._is_flatpak_repo():
-            cmd = ('flatpak', 'build-update-repo', self.path)
+            cmd = (
+                ['flatpak', 'build-update-repo'] +
+                sign_opts +
+                [str(self.path)]
+            )
         else:
-            cmd = ('ostree', f'--repo={self.path}', 'summary', '--update')
+            cmd = (
+                ['ostree', f'--repo={self.path}', 'summary', '--update'] +
+                sign_opts
+            )
         logger.info('Updating repo metadata with %s', ' '.join(cmd))
         subprocess.check_call(cmd)
 
-    def receive(self, refs, update_metadata=True, force=False, dry_run=False):
+    def update_repo_hook(self, refs):
+        """Run the configured update_hook
+
+        The program will be executed with the environment variable
+        OSTREE_RECEIVE_REPO set to the absolute path of the OSTree repository
+        and the environment variable OSTREE_RECEIVE_REFS set to the set of
+        refs received separated by whitespace.
+        """
+        if not self.config.update_hook:
+            raise OTReceiveConfigError('update_hook not set in configuration')
+
+        cmd = shlex.split(self.config.update_hook)
+        env = os.environ.copy()
+        env['OSTREE_RECEIVE_REPO'] = os.fspath(self.path.absolute())
+        env['OSTREE_RECEIVE_REFS'] = ' '.join(refs)
+
+        logger.info('Updating repo with %s', self.config.update_hook)
+        logger.debug('OSTREE_RECEIVE_REPO=%s', env['OSTREE_RECEIVE_REPO'])
+        logger.debug('OSTREE_RECEIVE_REFS=%s', env['OSTREE_RECEIVE_REFS'])
+        subprocess.check_call(cmd, env=env)
+
+    def receive(self, refs):
         # See what revisions we're pulling.
         _, remote_refs = self.remote_list_refs(self.REMOTE_NAME)
         if len(refs) == 0:
@@ -274,7 +471,7 @@ class OTReceiveRepo(OSTree.Repo):
                 raise OTReceiveError(
                     f'Could not find ref {ref} in summary file')
 
-            if force or remote_rev != current_rev:
+            if self.config.force or remote_rev != current_rev:
                 logger.debug('Pulling %s', ref)
                 refs_to_pull[ref] = remote_rev
 
@@ -322,7 +519,7 @@ class OTReceiveRepo(OSTree.Repo):
                                 'ref %s remote commit %s root equals %s',
                                 ref, remote_rev, current_rev
                             )
-                        if force:
+                        if self.config.force:
                             logger.info('Forcing merge of ref %s', ref)
                             refs_to_merge[ref] = remote_rev
 
@@ -332,14 +529,14 @@ class OTReceiveRepo(OSTree.Repo):
                 return set()
 
             # For a dry run, exit now before creating the refs
-            if dry_run:
+            if self.config.dry_run:
                 self.abort_transaction()
                 return refs_to_merge.keys()
 
             # Copy the pulled commits to the local ref so they get the
             # correct collection and ref bindings
             for ref, rev in refs_to_merge.items():
-                new_rev = self.copy_commit(rev, ref, force=force)
+                new_rev = self.copy_commit(rev, ref)
                 logger.debug('Set %s ref to %s', ref, new_rev)
 
             # All done, commit the changes
@@ -349,8 +546,11 @@ class OTReceiveRepo(OSTree.Repo):
             raise
 
         # Finally, regenerate the summary and metadata
-        if update_metadata:
-            self.update_repo_metadata()
+        if self.config.update:
+            if self.config.update_hook:
+                self.update_repo_hook(refs_to_merge.keys())
+            else:
+                self.update_repo_metadata()
 
         return refs_to_merge.keys()
 
@@ -358,13 +558,23 @@ class OTReceiveRepo(OSTree.Repo):
 class OTReceiveArgParser(ArgumentParser):
     """ArgumentParse for ostree-receive"""
     def __init__(self):
+        config_paths = ' or '.join(map(str, OTReceiveConfig.default_paths()))
         super().__init__(
             description='Pull from a remote repo to a dev repo',
+            epilog=(
+                'Many options can also be set in a config file '
+                f'({config_paths}). The config file uses YAML syntax and '
+                'must represent a YAML mapping.'
+            ),
+
+            # The global default is set to SUPPRESS so that options
+            # don't override config defaults.
+            argument_default=SUPPRESS,
         )
         self.add_argument('repo', metavar='REPO',
                           help='repository name to use')
         self.add_argument('url', metavar='URL', help='remote repository URL')
-        self.add_argument('refs', metavar='REF', nargs='*',
+        self.add_argument('refs', metavar='REF', nargs='*', default=None,
                           help='ostree refs to pull, all if none specified')
         self.add_argument('--no-update', dest='update', action='store_false',
                           help="""don't update repo metadata""")
@@ -373,12 +583,11 @@ class OTReceiveArgParser(ArgumentParser):
         self.add_argument('-f', '--force', action='store_true',
                           help=('force pull even if nothing changed or '
                                 'remote commits are not newer'))
-        self.set_defaults(log_level=logging.INFO)
         self.add_argument('-q', '--quiet', dest='log_level',
-                          action='store_const', const=logging.WARNING,
+                          action='store_const', const='WARNING',
                           help='disable most messages')
         self.add_argument('-v', '--verbose', dest='log_level',
-                          action='store_const', const=logging.DEBUG,
+                          action='store_const', const='DEBUG',
                           help='enable verbose messages')
         self.add_argument('--version', action='version',
                           version=f'%(prog)s {VERSION}')
@@ -387,11 +596,12 @@ class OTReceiveArgParser(ArgumentParser):
 def main():
     aparser = OTReceiveArgParser()
     args = aparser.parse_args()
+    config = OTReceiveConfig.load(args=args)
 
-    logging.basicConfig(level=args.log_level)
+    logging.basicConfig(level=config.log_level)
 
-    with OTReceiveRepo(args.repo, args.url) as repo:
-        repo.receive(args.refs, args.update, args.force, args.dry_run)
+    with OTReceiveRepo(args.repo, args.url, config) as repo:
+        repo.receive(args.refs)
 
 
 if __name__ == '__main__':
